/* File: main.c
 * Desc: Main sourcefile and program flow for the CardServices System of Zaphod 1
 * Auth: San Mehat, Woody
 * Note: 
 */

#include <termios.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <unistd.h>
#include <netinet/in.h>
#include <signal.h>
#include <errno.h>

#include <fcntl.h>
#include <sys/signal.h>
#include <sys/types.h>

#include <znodelist.h>
#include <afp.h>
#include <config.h>
#include <portfw.h>

#define BAUDRATE B19200					//38400
#define MODEMDEVICE "/dev/ttyS0"
#define _POSIX_SOURCE 1					/* POSIX compliant source */
#define MAX(a,b) ((a>b)?a:b)

#define ARE_U_THERE 0xC1
#define	NODE_STATUS 0xD9
#define BAD_LINK_PACKET 1

#define STALE_PACKET_TO	1
#define REQUEST_TO	20

#define	REPLY_TIMEOUT -1
#define	GOT_REPLY	-2
#define	REPLY_NO_DATA -3

#define	MAX_EEPROM_SIZE	512

#define EEPROM_BLANK	-4
#define EEPROM_ERROR	-5


#define deb(x) x

CONFIGURATION *Tconfigp_ptr;
Z_NODE *ZN_anchor = 0;
extern int errno;

void send_ll_nak (struct in_addr);
int send_afp_status_request (Z_NODE * Tnodep_ptr);
char *mytime ();
int process_link (int link_fd, int opcode, char* buff);
int	wait_response(int link_fd, int opcode, char* buff);
int	read_eeprom(int link_fd, char* data);
int write_eeprom(int link_fd, char* data, unsigned int count);
int	get_all_nodes(int link_fd);
int	process_node_status(char* cur_packet);
Z_NODE* get_node_ptr(int nodenum);


struct ee_header
{
char signature[4];	//0		'NWCF'
char ee_ip[4];			//4		head unit IP
char ee_netmask;		//8		zaphod netmask
char ee_version;		//9		0
char count[2];			//10	eeprom data size
char reserved[4];		//12	free
};


int I_flushflag = 0;

/*********************************************************************/

void 
handle_signal (int I_signo)
{
	switch (I_signo)
	{
	case SIGUSR1:
		{
			printf ("[%s] ***SIGUSR1 RECEIVED*** Reloading configuration data\n", mytime ());
			destroy_partitions (Tconfigp_ptr);
			memset (Tconfigp_ptr, 0, sizeof (CONFIGURATION));
			if (config_load (Tconfigp_ptr) < 0)
			{
				disp_label ("Unable to load configuration data");
				exit (1);
			}
			I_flushflag = 1;
			break;
		}
	}
	return;
}

/*********************************************************************/

int 
main (argc, argv)
		 char **argv;
{
	int I_sockfd;
	struct sockaddr_in serv_addr, cli_addr;
	void *Vp_buff;

	int link_fd, c, res, conf_fd;
	struct termios oldtio, newtio;
	char *eeprom_ptr;



	signal (SIGUSR1, handle_signal);

	system ("clear");


	if (!(eeprom_ptr = (char *) malloc (MAX_EEPROM_SIZE)))
	{
		disp_label ("Unable to allocate memory for eeprom data");
		exit (1);
	}


	/* Initialize serial comm system */
	disp_label ("Initializing facility link system");

	/* open the device to be non-blocking (read will return immediatly) */
	link_fd = open (MODEMDEVICE, O_RDWR | O_NOCTTY | O_NONBLOCK);
	if (link_fd < 0)
	{
		perror ("facility link");
		exit (1);
	}

	tcgetattr (link_fd, &oldtio);			/* save current port settings */
	/* set new port settings for non-canonical input processing */
	bzero (&newtio, sizeof (newtio));

	newtio.c_cflag = BAUDRATE | PARENB | CS8 | CLOCAL | CREAD;
	newtio.c_iflag = IGNBRK | INPCK | ICRNL;
	newtio.c_oflag = 0;
	newtio.c_lflag = 0;
	newtio.c_cc[VMIN] = 1;				// wait for up to 32 chars
	newtio.c_cc[VTIME] = 1;				// * 100 ms

	tcflush (link_fd, TCIFLUSH);
	tcsetattr (link_fd, TCSANOW, &newtio);

#if 0
	if (read_eeprom(link_fd, eeprom_ptr))
	{
		disp_label ("Unable to load eeprom data");
		exit (1);
	}

	conf_fd = open("zaphod.conf",O_CREAT);
	if (conf_fd <= 0)
	{
		disp_label ("Unable to create config file");
		exit (1);
	}
	res = write (conf_fd, eeprom_ptr + 0x10, *(unsigned int *)(eeprom_ptr + 0x0C));
	fclose (conf_fd);

	if (res != *(unsigned int *)(eeprom_ptr + 0x0C))
	{
		disp_label ("Error writing to config file");
		exit (1);
	}

#endif

	if(get_all_nodes(link_fd))
	{
		disp_label ("Unable to find node info");
		exit (1);
	}

	/* Retrieve configuration data */
	disp_label ("Loading Sub-system configuration data");
	if (!(Tconfigp_ptr = (CONFIGURATION *) malloc (sizeof (CONFIGURATION))))
	{
		disp_label ("Unable to allocate memory for configuration data");
		exit (1);
	}
	memset (Tconfigp_ptr, 0, sizeof (CONFIGURATION));
	if (config_load (Tconfigp_ptr) < 0)
	{
		disp_label ("Unable to load configuration data");
		exit (1);
	}
	/* Initialize socket system */
	disp_label ("Initializing socket system");
	if (!(I_sockfd = socket (AF_INET, SOCK_DGRAM, 0)) < 0)
	{
		perror ("socket");
		exit (1);
	}

	bzero ((char *) &serv_addr, sizeof (serv_addr));
	serv_addr.sin_family = AF_INET;
	serv_addr.sin_addr.s_addr = htonl (INADDR_ANY);
	serv_addr.sin_port = htons (6442);

	if (bind (I_sockfd, (struct sockaddr *) &serv_addr, sizeof (serv_addr)) < 0)
	{
		perror ("bind");
		exit (1);
	}

	disp_label ("Socket bound. Setting up for async i/o");

	if (!(Vp_buff = (void *) malloc (MAX_AFPSIZE)))
	{
		disp_label ("Unable to allocate memory for async input buffer");
		exit (1);
	}
	printf ("[%s] Allocated %d bytes at 0x%x for async input buffer\n", mytime (), MAX_AFPSIZE, Vp_buff);



	while (1)
	{
		struct timeval to;
		int I_rc;
		int	maxfd;
		fd_set readfds;

		to.tv_sec = 0;
		to.tv_usec = 5000;
		FD_ZERO (&readfds);
		FD_SET (I_sockfd, &readfds);
		FD_SET (link_fd, &readfds);
		maxfd = MAX (I_sockfd, link_fd) + 1;

		if ((I_rc = select (maxfd, &readfds, 0, 0, &to)) < 0)
		{
			if (errno = EINTR)
				continue;								/* We were interrupted by a signal... continue james... */
			perror ("select");
			exit (1);
		}
		else if (!I_rc)
		{
			/* Check to see if the nodes need to be reconfigured */
			if (I_flushflag)
			{
				reset_all_nodes ();
				I_flushflag = 0;
			}
//      update_spinner();
			//      disp_label("Performing periodic processing of event queue");
			I_rc = process_event_queue ();
			if (I_rc > 0)
				printf ("[%s] %d outbound events processed while waiting for inbound data\n", mytime (), I_rc);
			continue;
		}
		else if (FD_ISSET (link_fd, &readfds))
		{
			/* process link data */
			if (process_link (link_fd, 0, 0) < 0)
			{
				perror ("bad facility link");
				exit (1);
			}
		}
		else if (FD_ISSET (I_sockfd, &readfds))
		{
			int cli_len = sizeof (cli_addr);
			Z_NODE *ZN_nodeptr;
			int I_clinodenum;

			if ((I_rc = recvfrom (I_sockfd, Vp_buff, MAX_AFPSIZE, 0, &cli_addr, &cli_len)) < 0)
			{
				perror ("recvfrom");
				exit (1);
			}
			if (!I_rc)
			{
				disp_label ("Ignoring rc of 0 from read()");
				continue;
			}
			I_clinodenum = ((cli_addr.sin_addr.s_addr & 0xff000000) >> 24);
#if 0
			printf ("[%s] Received %d bytes from slave %d (%s)\n",
							mytime (),
							I_rc,
							I_clinodenum,
							inet_ntoa (cli_addr.sin_addr));
#endif

#if 0
			/* 
			   If we're receiving messages from a slave... 
			   we should make sure he has a nodelist entry 
			 */
			if (!ZN_anchor)
			{
				disp_label ("Creating nodelist");
				if (!(ZN_anchor = (Z_NODE *) malloc (sizeof (Z_NODE))))
				{
					disp_label ("Unable to allocate memory for first nodelist entry");
					send_ll_nak (cli_addr.sin_addr);	/* send low level nak to client */
					continue;
				}
				memset (ZN_anchor, 0, sizeof (Z_NODE));
				ZN_anchor->I_nodenum = I_clinodenum;
				ZN_anchor->flags = FLAG_POWR | FLAG_INST;
				ZN_anchor->phase.I_operation = PHASE_POWERUP;		/* XXX: hacked here */
				ZN_anchor->phase.UI_alarmtime = (((unsigned int) time (NULL)) + PHASE_POWERUP_TIMEOUT);
				ZN_anchor->UI_ipaddr = cli_addr.sin_addr.s_addr;
				ZN_nodeptr = ZN_anchor;
			}
			else
#endif
			{
				for (ZN_nodeptr = ZN_anchor;
						 ZN_nodeptr; ZN_nodeptr = ZN_nodeptr->next)
				{
					if (ZN_nodeptr->I_nodenum == I_clinodenum)
						break;
				}
				if (!ZN_nodeptr)
				{
					/* should not happen - we did it in node_status phase! */
					Z_NODE *ZN_link;

					printf ("[%s] Adding UNKNOWN node %d to nodelist\n",
					 mytime (), I_clinodenum);
					if (!(ZN_nodeptr = (Z_NODE *) malloc (sizeof (Z_NODE))))
					{
						disp_label ("Unable to allocate memory for the nodelist entry");
						send_ll_nak (cli_addr.sin_addr);	/* send low level nak to client */
						continue;
					}
					memset (ZN_nodeptr, 0, sizeof (Z_NODE));
					ZN_nodeptr->I_nodenum = I_clinodenum;
					ZN_nodeptr->flags = FLAG_POWR | FLAG_INST;
					ZN_nodeptr->UI_ipaddr = cli_addr.sin_addr.s_addr;
					ZN_nodeptr->phase.I_operation = PHASE_POWERUP;	/* XXX: hacked here */
					ZN_nodeptr->phase.UI_alarmtime = (((unsigned int) time (NULL)) + PHASE_POWERUP_TIMEOUT);
					for (ZN_link = ZN_anchor; ZN_link->next; ZN_link = ZN_link->next);
					ZN_link->next = ZN_nodeptr;
				}
			}

			/* Now call packet processor */
			if (process_packet (Vp_buff, ZN_nodeptr) < 0)
			{
				disp_label ("process_packet() requested shutdown");
				free (Vp_buff);
				close (I_sockfd);


				free (Tconfigp_ptr);
				free (eeprom_ptr);

				/* restore old port settings */
				tcsetattr (link_fd, TCSANOW, &oldtio);
				close (link_fd);

				exit (0);
			}
			I_rc = process_event_queue ();
			if (I_rc > 0)
				printf ("[%s] %d outbound events processed after processing inbound data\n", mytime (), I_rc);
			continue;
		}
		else
		{
			disp_label ("select() assert! failure");
			exit (1);
		}
	}
}

/*********************************************************************/

/* Run through nodelist, check for any nodes which should be sent AFP requests... do so */
int 
process_event_queue ()
{
	Z_NODE *Tnodep_ptr = ZN_anchor;
	unsigned int UI_currenttime = (int) time (NULL);
	unsigned int I_numprocessed = 0;

	/* check for inserted or removed cards */
	/* if card inserted, power it up and set phase to PHASE_POWER */
	/* Run through nodelist and do the AFP request thing .. god i hate mono... */
	while (Tnodep_ptr)
	{
		if (Tnodep_ptr->hwflags & HW_CHANGE_FLAG)
		{
			Tnodep_ptr->hwflags &= ~HW_CHANGE_FLAG;	//mark change noticed

			switch (Tnodep_ptr->hwflags)
			{
				case HW_NO_NODE:				/* Node not installed */
				{
					/* was node just unplugged? Some bits in flags would have been set.*/
					if (Tnodep_ptr->flags)
					{
						power (Tnodep_ptr, 0);
						Tnodep_ptr->hwflags = 0;
						Tnodep_ptr->flags = 0;
						Tnodep_ptr->phase.I_operation = PHASE_NO_POWER;
						Tnodep_ptr->phase.UI_alarmtime = UI_currenttime + PHASE_NO_POWER_TIMEOUT;
					}
					else
					{
						deb(printf("event_queue: ignoring NO_NODE status for node %d.\n",
									Tnodep_ptr->I_nodenum));
					}
					break;
				}

				case HW_NODE_OK:				/* Node responding OK to the MPU */

					printf("Node %d OK.\n",Tnodep_ptr->I_nodenum);
					if (!Tnodep_ptr->flags)
					{
						

					}
					break;

				case HW_NODE_POWER_OFF:	/* Node has no power */
					if (!Tnodep_ptr->flags)	/* we are just starting */
					{
						printf("Powering node %d up.\n",Tnodep_ptr->I_nodenum);
					}
					else
					{
						printf("Node %d lost power! Restoring...\n",Tnodep_ptr->I_nodenum);
						power (Tnodep_ptr, 0);
					}
						power (Tnodep_ptr, 1);
						Tnodep_ptr->phase.I_operation = PHASE_POWERUP;
						Tnodep_ptr->phase.UI_alarmtime = UI_currenttime + PHASE_POWERUP_TIMEOUT;

					break;

				case HW_NODE_DEAD:			/* Node not responding to the MPU */

				default:
						/* we should actually update the status now - it is not valid!!! */
						(printf("event_queue: ignoring 0x%X status for node %d.\n",
									Tnodep_ptr->hwflags, Tnodep_ptr->I_nodenum));

			}
		}

		/* Check phase alarm clock */
		if (Tnodep_ptr->phase.UI_alarmtime <= UI_currenttime)
		{
			/* Something alarmed.. */
			switch (Tnodep_ptr->phase.I_operation)
			{
			case PHASE_POWERUP:			/* Time expired for node to request configuration */
				{
					I_numprocessed++;
					printf ("[%s] Node %d powerup time exceeded\n", mytime (), Tnodep_ptr->I_nodenum);
					Tnodep_ptr->UI_failcount++;
					if (Tnodep_ptr->UI_failcount >= 3)
					{
						printf ("[%s] Node %d powerup retry limit exceeded\n", mytime (), Tnodep_ptr->I_nodenum);
						Tnodep_ptr->phase.I_operation = PHASE_FAULT;
						Tnodep_ptr->phase.UI_alarmtime = UI_currenttime + PHASE_FAULT_TIMEOUT;
					}
					else
					{
						printf ("[%s] Node %d is being power cycled\n", mytime (), Tnodep_ptr->I_nodenum);
						power (Tnodep_ptr, 0);
						power (Tnodep_ptr, 1);
						Tnodep_ptr->phase.I_operation = PHASE_POWERUP;
						Tnodep_ptr->phase.UI_alarmtime = UI_currenttime + PHASE_POWERUP_TIMEOUT;
					}
					break;
				}
			case PHASE_CONFIG:				/* Time expired for node to enter ready state     */
				{
					I_numprocessed++;
					printf ("[%s] Node %d readyness time exceeded\n", mytime (), Tnodep_ptr->I_nodenum);
					Tnodep_ptr->UI_failcount++;
					if (Tnodep_ptr->UI_failcount >= 3)
					{
						printf ("[%s] Node %d readyness retry limit exceeded\n", mytime (), Tnodep_ptr->I_nodenum);
						Tnodep_ptr->phase.I_operation = PHASE_FAULT;
						Tnodep_ptr->phase.UI_alarmtime = UI_currenttime + PHASE_FAULT_TIMEOUT;
					}
					else
					{
						printf ("[%s] Node %d is being power cycled\n", mytime (), Tnodep_ptr->I_nodenum);
						power (Tnodep_ptr, 0);
						power (Tnodep_ptr, 1);
						Tnodep_ptr->phase.I_operation = PHASE_POWERUP;
						Tnodep_ptr->phase.UI_alarmtime = UI_currenttime + PHASE_POWERUP_TIMEOUT;
					}
					break;
				}
			case PHASE_IDLE:					/* Time for another AFP request                   */
				{
					I_numprocessed++;
					printf ("[%s] Node %d status dispatch alarm\n", mytime (), Tnodep_ptr->I_nodenum);
					send_afp_status_request (Tnodep_ptr);
					Tnodep_ptr->phase.I_operation = PHASE_STATUS;
					Tnodep_ptr->phase.UI_alarmtime = UI_currenttime + PHASE_STATUS_TIMEOUT;
					break;
				}
			case PHASE_STATUS:				/* Time expired for AFP STATUS response           */
				{
					I_numprocessed++;
					printf ("[%s] Node %d status time exceeded (IP_PORTFW OFF)\n", mytime (), Tnodep_ptr->I_nodenum);
					/*ip_portfw_remove (Tnodep_ptr);TEMP*/
					Tnodep_ptr->UI_failcount++;
					if (Tnodep_ptr->UI_failcount >= 3)
					{
						printf ("[%s] Node %d is being soft reset\n", mytime (), Tnodep_ptr->I_nodenum);
						send_soft_reset (Tnodep_ptr);
						Tnodep_ptr->phase.I_operation = PHASE_RESET;
						Tnodep_ptr->phase.UI_alarmtime = UI_currenttime + PHASE_RESET_TIMEOUT;
					}
					else
					{
						printf ("[%s] Node %d status retry\n", mytime (), Tnodep_ptr->I_nodenum);
						send_afp_status_request (Tnodep_ptr);		/* Send another AFP request */
						Tnodep_ptr->phase.I_operation = PHASE_STATUS;
						Tnodep_ptr->phase.UI_alarmtime = UI_currenttime + PHASE_STATUS_TIMEOUT;
					}
					break;
				}
			case PHASE_RESET:				/* Time expired for node to reset and request configuration */
				{
					I_numprocessed++;
					printf ("[%s] Node %d reset time exceeded\n", mytime (), Tnodep_ptr->I_nodenum);
					Tnodep_ptr->phase.I_operation = PHASE_FAULT;
					Tnodep_ptr->phase.UI_alarmtime = UI_currenttime + PHASE_FAULT_TIMEOUT;
					break;
				}
			case PHASE_FAULT:				/* Time for another RESET message                 */
				{
					I_numprocessed++;
					printf ("[%s] Node %d resurrection alarm\n", mytime (), Tnodep_ptr->I_nodenum);
					power (Tnodep_ptr, 0);
					power (Tnodep_ptr, 1);
					Tnodep_ptr->phase.I_operation = PHASE_FAULT;
					Tnodep_ptr->phase.UI_alarmtime = UI_currenttime + PHASE_FAULT_TIMEOUT;
					break;
				}
			default:
				{
					printf ("[%s] Node %d is in unknown phase %d\n", mytime (), Tnodep_ptr->phase.I_operation);
					break;
				}
			}
		}
		Tnodep_ptr = Tnodep_ptr->next;
	}
	return (I_numprocessed);
}

/*********************************************************************/

int 
send_afp_status_request (Z_NODE * Tnodep_ptr)
{
	unsigned int UI_currenttime = (int) time (NULL);
	AFP_HDR pkthdr;

	Tnodep_ptr->flags |= FLAG_SENT;
	Tnodep_ptr->UI_afptstamp = UI_currenttime;

	pkthdr.UI_txid = random ();
	pkthdr.UI_cmd = REQ_STATUS;
	pkthdr.UI_dlen = 0;
	if (send_data (Tnodep_ptr, &pkthdr, sizeof (pkthdr)) < 0)
	{
		printf ("[%s] Unable to send data\n", mytime ());
		return (-1);
	}
	return (0);
}

/*********************************************************************/

int 
send_soft_reset (Z_NODE * Tnodep_ptr)
{
	AFP_HDR pkthdr;


	pkthdr.UI_txid = random ();
	pkthdr.UI_cmd = REQ_RESET;
	pkthdr.UI_dlen = 0;
	if (send_data (Tnodep_ptr, &pkthdr, sizeof (pkthdr)) < 0)
	{
		printf ("[%s] Unable to send data\n", mytime ());
		return (-1);
	}
	return (0);
}

/*********************************************************************/

int 
reset_all_nodes ()
{
	Z_NODE *Tnodep_ptr = ZN_anchor;

	while (Tnodep_ptr)
	{
		send_soft_reset (Tnodep_ptr);
		Tnodep_ptr->phase.I_operation = PHASE_RESET;
		Tnodep_ptr->phase.UI_alarmtime = (((unsigned int) time (NULL)) + PHASE_RESET_TIMEOUT);
		Tnodep_ptr->flags = 0;
		Tnodep_ptr->hwflags = 0;
		Tnodep_ptr = Tnodep_ptr->next;
	}
	return (0);
}

/*********************************************************************/

disp_label (char *msg)
{
	printf ("[%s] %s\n", mytime (), msg);
}

/*********************************************************************/

/* 
   Function to create a new socket and send a low level NAK to the client.
   Note theres weird handling because we cant trust that we have a good txid...
   so we just send a global NAK to the client and leave it up to him to remember
   what transaction we're talking about
 */

void 
send_ll_nak (struct in_addr addr)
{

	int I_sockfd;
	struct sockaddr_in sock_addr;
	AFP_HDR pkt;

	bzero ((char *) &sock_addr, sizeof (sock_addr));
	sock_addr.sin_family = AF_INET;
	sock_addr.sin_addr = addr;
	sock_addr.sin_port = htons (6443);

	if ((I_sockfd = socket (AF_INET, SOCK_DGRAM, 0)) < 0)
	{
		perror ("socket");
		printf ("[CardServices] Unable to send low level NAK to slave at %s (%d)\n", inet_ntoa (addr), I_sockfd);
		return;
	}
	pkt.UI_txid = 0;
	pkt.UI_cmd = NAK;
	if (sendto (I_sockfd, (void *) &pkt, sizeof (pkt), 0, (struct sockaddr *) &sock_addr, sizeof (sock_addr)) < 0)
	{
		perror ("sendto");
		printf ("[CardServices] Unable to send low level NAK to slave at %s (%d)\n", inet_ntoa (addr), I_sockfd);
	}
	close (I_sockfd);
	return;
}

/*********************************************************************/

/* Debug thing for timestamps */
char timebuff[64];
char *
mytime ()
{
	time_t timeval;

	timeval = time (NULL);
	strcpy (timebuff, ctime (&timeval));
	timebuff[(strlen (timebuff) - 1)] = 0;
	return (timebuff);
}

/*********************************************************************/

int I_last = 0;
char spinnything[] =
{'|', '/', '-', '\\'};
int 
update_spinner ()
{

	printf ("%c%c", spinnything[I_last], (char) 8);
	fflush (stdout);
	I_last++;
	if (I_last == 4)
		I_last = 0;
}

/*********************************************************************/

int 
power (Z_NODE * Tnodep_ptr, int I_onoff)
{

	if (I_onoff == 0)
		printf ("[%s] Node %d POWER OFF\n", mytime (), Tnodep_ptr->I_nodenum);
	else
		printf ("[%s] Node %d POWER ON\n", mytime (), Tnodep_ptr->I_nodenum);
}

/*********************************************************************/

nat_remove (Z_NODE * Tnodep_ptr)
{
	printf ("[%s] Node %d removed from NAT tables\n", mytime (), Tnodep_ptr->I_nodenum);
}

/*********************************************************************/

nat_add (Z_NODE * Tnodep_ptr)
{
	printf ("[%s] Node %d added to NAT tables\n", mytime (), Tnodep_ptr->I_nodenum);
}

/*********************************************************************/

int	checksum(char *packet)
{
	return(0);
}

/*********************************************************************/

int	build_checksum(char *packet)
{
	return(0xABCD);
}

/*********************************************************************/

/*
	This routine is normally called from the select() semaphore.
	In such case it will read whatever is waiting in COM2 port, and if the
	data can be assembled into a packet - process it.

	Another way of calling it is from a wait_reply() routine.
	In this case it may have no data - return REPLY_NO_DATA


	Return: -3 == no data
					-2 == requested packet ready
					
					len== number of bytes in the buffer
						0== processed OK
*/

int process_link (int link_fd, int opcode, char* buff)
{
static char packet_buf[128];
static int pack_size;
static unsigned int pack_time;

char packet[128];
char reply[16];
char *cur_packet;
int	pack_err;
int pack_code;
int i,j,k;

	i = read (link_fd, packet, 128);

	/* if called without select - if no data - just return...*/
	if (i == -1)
	{
//		deb(printf("link: return NO_DATA (%d)...\n",REPLY_NO_DATA));
		return(REPLY_NO_DATA);
	}
#if 1
	deb(printf("process_link[%d]: ",i));
	for (j = 0; j < i; j++)
	{
		deb(printf("%02X ",packet[j]));
	}
	deb(printf("\n"));
#endif

	/* by default - try to read from the in buff. */
	cur_packet = packet;

	/* we have that many bytes waiting...*/
	if (0 != pack_size)
	{
		j = (int) time (NULL);
		if ((unsigned int)j > pack_time + STALE_PACKET_TO)
		{
			/* invalidate old bytes waiting - most likely garbage... */
			deb(printf("process_link(): stale packet(%d byte(s))...\n", pack_size));
			pack_size = 0;
		}
		else
		{
			deb(printf("process_link(): append to packet[%d]...\n", pack_size));
			cur_packet = packet_buf;


			for (k = 0; k < i; k++)
			{
				cur_packet[pack_size++] = packet[k];
			}

			i = pack_size;		//fake receiving that many bytes
		}
	}

	/* sanity check: we do not expect all 128 bytes back, and pack size needs
		to be in range as well	*/
	if ((i == 128) || cur_packet[0] > 127)
	{
		/* fake some garbage short packet with a perfectly invalid checksum */
		cur_packet[0] = 0x04;
		cur_packet[1] = 0xFF;
		cur_packet[2] = 0x00;
		cur_packet[3] = 0x00;
	}

	if (i < cur_packet[0])	//did we receive full packet?
	{
		if(cur_packet != packet_buf)
		{
			for (j = 0; j < i; j++)
			{
				packet_buf[j] = cur_packet[j];
			}
		pack_size = i;
		pack_time = (int) time (NULL);
		}

		return(pack_size);				//no - wait for more...
	}

	/* we have the packet ready! Clear the incomplete packet ptr.*/
	pack_size = 0;
	if (opcode && (opcode == cur_packet[1]))
	{
		/* we got the reply, we were waiting for... */
		if (buff)
		{
			for (j = 0; j < i; j++)
			{
				buff[j] = cur_packet[j];
			}
		}
		return(GOT_REPLY);		//this wait is over...
	}

	/* else it is unsolicited message from the MPU, so process and reply */
	{
		switch (*(cur_packet+1))	//work with the command code...
		{

			case ARE_U_THERE:
				if ((*cur_packet != 0x06) || (checksum(cur_packet)))
				{
					pack_err = BAD_LINK_PACKET;
					break;
				}
				reply[0] = 0x06;

				reply[1] = 0xAA;		//pack_code
				reply[2] = 0x00;		//pack_err == no error

				reply[3] = packet[0x02] ^ packet[0x03];

				k = build_checksum(reply);
				reply[4] = k & 0xFF;
				reply[5] = k >> 8;

				pack_err = 0;
				break;


			case NODE_STATUS:
				if (checksum(cur_packet))
				{
					pack_code = 0xFE;
					pack_err = BAD_LINK_PACKET;
					break;
				}

				/* save the node info... */
				k = process_node_status(cur_packet);

				/* no reply required */
				return (0);

			default:
				(printf("process_link(): unexpected packet (cmd = 0x%X).\n",
				 cur_packet[1]));
				pack_code = 0xFF;
				pack_err = BAD_LINK_PACKET;
				break;
		}


		if (pack_err)
		{
				deb(printf("process_link(): returning error for cmd = 0x%X.\n", cur_packet[1]));

				reply[0] = 0x05;
				reply[1] = pack_code;
				reply[2] = pack_err;

				k = build_checksum(reply);
				reply[3] = (char)(k & 0xFF);
				reply[4] = (char)(k >> 8);
		}

		/* send the reply */
		j = write (link_fd, reply, reply[0]);

		return (0);

	}
}

/*********************************************************************/

int	wait_response(int link_fd, int opcode, char* buff)
{
	unsigned int	start_time = (int) time (NULL);
	int	rc;
	deb(printf("wait_response for %X.\n",opcode));
	while (((rc = process_link(link_fd, opcode, buff)) != GOT_REPLY) && 
				(start_time + REQUEST_TO > (int) time (NULL)))
	{
#if 0
		if (buff)
		{
			process_event_queue();
		}
#endif
	}
	//deb(printf("wait_response() rc=%d.\n",rc));

	return(rc);
}

/*********************************************************************/

int read_eeprom(int link_fd, char* eeprom_ptr)
{
	int	i,j,k;
	char	cmdbuff[8];
	char	buff[128];
	unsigned int start_time = (int) time (NULL);
	unsigned int ee_offset;
	volatile unsigned int ee_size = MAX_EEPROM_SIZE;
	char* pdata;

	/* wait for a MPU ping packet... */
	printf("Waiting for the first MPU packet...\n");

	while((i = process_link (link_fd, 0, NULL)) &&
				(start_time + REQUEST_TO > (int) time (NULL)))
	{
	};

	/* if timeout getting any packet from MPU - error */
	if (i)
	{
		printf("MPU not responding...\n",i);
		return(i);
	}

	cmdbuff[0] = 0x07;	//len
	cmdbuff[1] = 0xE2;	//opcode
	cmdbuff[2] = 120;		//data len
	cmdbuff[3] = 0x0;		//offset lo
	cmdbuff[4] = 0x0;		//offset hi
	cmdbuff[5] = 0x61;	//checksum
	cmdbuff[6] = 0x62;

	ee_offset = 0;

	printf("Reading EEPROM settings (%d bytes)...\n",ee_size);
	while(ee_offset < ee_size)
	{

		/* if less data then 120 bytes required - change the size field */
		if ((i = ee_size - ee_offset) < 120)
		{
			cmdbuff[2] = (char)(i & 0xFF);
		}

		cmdbuff[3] = (char)(ee_offset & 0xFF);
		cmdbuff[4] = (char)(ee_offset >> 8);

		i = build_checksum(cmdbuff);
		cmdbuff[5] = (char)(i & 0xFF);
		cmdbuff[6] = (char)(i >> 8);

		deb(printf("read_eeprom(): len=%d, offset=%d.\n",cmdbuff[2],ee_offset));

		start_time = (int) time (NULL);
		process_link (link_fd, 0, NULL);		//flush the line

		i = wait_response(link_fd, ARE_U_THERE, NULL);

		if (i == GOT_REPLY)
		{
			write (link_fd, cmdbuff, 7);
			i = wait_response(link_fd, 0xA8, buff);
			if (i != GOT_REPLY)
			{
				write (link_fd, cmdbuff, 7);
				i = wait_response(link_fd, 0xA8, buff);
			}

			if (i == GOT_REPLY)
			{
				j = buff[0] - 5;
				pdata = &buff[3];

#if 1
				if (ee_offset == 0)	//very first packet???
				{
					/* expect the header struct ee_header... */
					// check signature word
					if((*pdata == 'N') && (*(pdata+1) == 'W') &&
						 (*(pdata+2) == 'C') && (*(pdata+3) == 'F'))
					{
						ee_size = *(pdata+10) + (*(pdata+11) << 8);

						ee_size += 16;	//sizeof(ee_header) 

						if (ee_size > MAX_EEPROM_SIZE)
							return(EEPROM_ERROR);

						//adjust the pointer/count to skip the header...
						ee_offset = 16;	//sizeof(ee_header)
						j -= 16;				//sizeof(ee_header)
					}
					else
					{
						return(EEPROM_BLANK);
					}

				}
#endif
				deb(printf("read_eeprom(): received %d bytes.\n",j));
				for (k=0; k < j; k++)
				{
					*(eeprom_ptr + ee_offset) = buff[k];
					ee_offset++;
				}
			}
		}

		if(i != GOT_REPLY)
			return(i);
	}

	return(0);
}

/*********************************************************************/

int write_eeprom(int link_fd, char* eeprom_ptr, unsigned int ee_size)
{
	int	i,j,k;
	char	replybuf[8];
	char	cmdbuff[128];
	unsigned int start_time = (int) time (NULL);
	unsigned int ee_offset;
//	unsigned int count;
	char* pdata;

	if (ee_size >= MAX_EEPROM_SIZE)
	{
		return(EEPROM_ERROR);
	}

	/* wait for a MPU ping packet... */
	printf("Waiting for the MPU ping packet...\n");

	while((i = process_link (link_fd, 0, NULL)) &&
				(start_time + REQUEST_TO > (int) time (NULL)))
	{
	};

	/* if timeout getting any packet from MPU - error */
	if (i)
	{
		printf("MPU not responding...\n",i);
		return(i);
	}

	ee_offset = 16;		//sizeof (ee_struct) == start after the header

	cmdbuff[0] = 127;	//len
	cmdbuff[1] = 0xF3;	//opcode
	cmdbuff[2] = 120;		//data len


	printf("Writing EEPROM settings (%d bytes)...\n",ee_size);
	while(ee_offset < ee_size)
	{
		/* if less data then 120 bytes to send - change the size field */
		if ((i = ee_size - ee_offset) < 120)
		{
			cmdbuff[2] = (char)(i & 0xFF);			//data length
			cmdbuff[0] = (char)(i & 0xFF) + 7;	//packet length
		}

		cmdbuff[3] = (char)(ee_offset & 0xFF);
		cmdbuff[4] = (char)(ee_offset >> 8);

		for(j = 0; j < i; j++)
		{
			cmdbuff[j+5] = *eeprom_ptr++;		//	 + ee_offset + j;
		}

//		cmdbuff[0] = i + 7;

		j = build_checksum(cmdbuff);
		cmdbuff[i+5] = (char)(j & 0xFF);
		cmdbuff[i+6] = (char)(j >> 8);

		deb(printf("write_eeprom(): len=%d, offset=%d.\n",cmdbuff[2],ee_offset));

		start_time = (int) time (NULL);
		process_link (link_fd, 0, NULL);		//flush the line

		j = wait_response(link_fd, ARE_U_THERE, NULL);

		if (j == GOT_REPLY)
		{
			write (link_fd, cmdbuff, i + 7);
			j = wait_response(link_fd, 0xA9, replybuf);
			if (j != GOT_REPLY)
			{
				write (link_fd, cmdbuff, i + 7);
				j = wait_response(link_fd, 0xA9, replybuf);
			}

			if (j == GOT_REPLY)
			{
				//if written less then requested or error reported - return...
				if ((i != replybuf[2]) || (0 != replybuf[3]))
					return(EEPROM_ERROR);

			}
			else
				return(EEPROM_ERROR);

		}
		ee_offset += i;
	}

#if 1
	// at the very end - write the header...
	printf("Writing EEPROM header (16 bytes)...\n");

	cmdbuff[0] = 23;		//len = 7 + 16 bytes
	cmdbuff[1] = 0xF3;	//opcode
	cmdbuff[2] = 16;		//data len
	cmdbuff[3] = 0x0;		//offset lo
	cmdbuff[4] = 0x0;		//offset hi

	pdata = &cmdbuff[5];	//create the ee_header struct here

	*(pdata)   == 'N';
	*(pdata+1) == 'W';
	*(pdata+2) == 'C';
	*(pdata+3) == 'F';

	//skip for now the IP stuff
	*(pdata+4) == 0;
	*(pdata+5) == 0;
	*(pdata+6) == 0;
	*(pdata+7) == 0;

	//skip netmask
	*(pdata+8) == 0;

	//default header rev
	*(pdata+9) == 0;

	*(pdata+10)== ee_size & 0xFF;
	*(pdata+11)== (ee_size >> 8) & 0xFF;

	j = build_checksum(cmdbuff);
	cmdbuff[i+21] = (char)(j & 0xFF);
	cmdbuff[i+22] = (char)(j >> 8);

	start_time = (int) time (NULL);
	process_link (link_fd, 0, NULL);		//flush the line

	j = wait_response(link_fd, ARE_U_THERE, NULL);

	if (j == GOT_REPLY)
	{
		write (link_fd, cmdbuff, 23);
		j = wait_response(link_fd, 0xA9, replybuf);
		if (j != GOT_REPLY)
		{
			write (link_fd, cmdbuff, 23);
			j = wait_response(link_fd, 0xA9, replybuf);
		}

		if (j == GOT_REPLY)
		{
			//if written less then requested or error reported - return...
			if ((i != replybuf[2]) || (0 != replybuf[3]))
				return(EEPROM_ERROR);
		}
		else
			return(EEPROM_ERROR);
	}
#endif
	return(0);
}

/*********************************************************************/

int	process_node_status(char* cur_packet)
{
Z_NODE *Tnodep_ptr = ZN_anchor;
int i,j,k;

	j = 3;									/* point  to the first data byte */
	k = cur_packet[0] - 5;	/* number of status bytes in the packet */

	while (k)
	{

		i = cur_packet[j];
		i &= 0x3F;					/* we have 6 bits for the node number */


		/* find the right node (or create a new one) */
		if (Tnodep_ptr = get_node_ptr(i))
		{
				Tnodep_ptr->hwflags = HW_CHANGE_FLAG + (cur_packet[j]>>6);
		}
		else
		{

			(printf("Error processing the node status message. [0x%X].\n",
				 cur_packet[j]));
			return(1);
		}
		j++;
		k--;
	}
	return(0);
}


/*********************************************************************/

Z_NODE* get_node_ptr(int nodenum)
{
Z_NODE *ZN_link;
Z_NODE *ZN_nodeptr;

	if (ZN_anchor)
	{
		for (ZN_nodeptr = ZN_anchor;
				 ZN_nodeptr; ZN_nodeptr = ZN_nodeptr->next)
		{
			if (ZN_nodeptr->I_nodenum == nodenum)
				return (ZN_nodeptr);
		}
	}

	//else either no ZN_anchor, or we need to add a new node...

	if (!ZN_anchor)
		disp_label ("Creating nodelist");

	if (!(ZN_nodeptr = (Z_NODE *) malloc (sizeof (Z_NODE))))
	{
		disp_label ("Unable to allocate memory for the node entry.");
		return (NULL);
	}

	printf ("[%s] Adding node %d to nodelist\n", mytime (), nodenum);
	
	memset (ZN_nodeptr, 0, sizeof (Z_NODE));
	ZN_nodeptr->I_nodenum = nodenum;
	ZN_nodeptr->phase.I_operation = PHASE_NO_POWER;
	ZN_nodeptr->phase.UI_alarmtime = (((unsigned int) time (NULL)) + PHASE_NO_POWER_TIMEOUT);

	if (!ZN_anchor)
	{
		ZN_anchor = ZN_nodeptr;
	}
	else
	{

		for (ZN_link = ZN_anchor; ZN_link->next; ZN_link = ZN_link->next);
		ZN_link->next = ZN_nodeptr;
	}

	return(ZN_nodeptr);
}

/*********************************************************************/

int get_all_nodes(link_fd)
{
	int	i;

	char	cmdbuff[4];
	char	buff[128];

	cmdbuff[0] = 0x07;	//len
	cmdbuff[1] = 0xED;	//opcode
	cmdbuff[2] = 0x43;	//checksum
	cmdbuff[3] = 0x44;

	printf("Requesting node information...\n");
	i = wait_response(link_fd, ARE_U_THERE, NULL);

	if (i == GOT_REPLY)
	{
		write (link_fd, cmdbuff, 4);
		i = wait_response(link_fd, 0xD9, buff);

		if (i == GOT_REPLY)
		{
			return(process_node_status(buff));
		}
		else
		{
			return(1);
		}
	}
}

/*********************************************************************/


ip_portfw_setsockopt (int cmd, struct ip_portfw_edits *data, int length)
{
	static int sockfd = -1;
	int ret;

	if (sockfd == -1)
	{
		if ((sockfd = socket (AF_INET, SOCK_RAW, IPPROTO_RAW)) == -1)
		{
			perror("ip_portfw: socket creation failed");
			exit(1);
		}
	}
	ret = setsockopt (sockfd, IPPROTO_IP, cmd, (char*) data, length);
	if (ret)
		perror ("ip_portfw: setsockopt failed");
}

ip_portfw_add (__u16 protocol, __u32 laddr, __u16 lport, __u32 raddr, __u16 rport, short weight, int nodenum)
{
	struct ip_portfw_edits add_serv;
	add_serv.protocol = protocol;
	add_serv.laddr = laddr;
	add_serv.lport = lport;
	add_serv.raddr = raddr;
	add_serv.rport = rport;
	add_serv.weight = weight;
	ip_portfw_setsockopt (IP_PORTFW_ADD, &add_serv, sizeof(add_serv));
	printf ("[%s] Node %d added to IP_PORTFW tables\n", mytime (), nodenum);
}

ip_portfw_remove (__u16 protocol, __u32 laddr, __u16 lport, __u32 raddr, __u16 rport, short weight, int nodenum)
{
	struct ip_portfw_edits del_serv;
	del_serv.protocol = protocol;
	del_serv.laddr = laddr;
	del_serv.lport = lport;
	del_serv.raddr = raddr;
	del_serv.rport = rport;
	del_serv.weight = weight;
	ip_portfw_setsockopt (IP_PORTFW_DEL, &del_serv, sizeof(del_serv));
	printf ("[%s] Node %d removed from IP_PORTFW tables\n", mytime (), nodenum);
}

ip_portfw_bind_lbm(__u16 protocol, __u32 laddr, __u16 lport, char *lbm_name)
{
	struct ip_portfw_edits bind_serv;
	bind_serv.protocol = protocol;
	bind_serv.laddr = laddr;
	bind_serv.lport = lport;
	strncpy(lbm_name, bind_serv.lbm_name, IP_PORTFW_LBM_NAME_LEN);
	ip_portfw_setsockopt (IP_PORTFW_BIND, &bind_serv, sizeof(bind_serv));
	printf ("[%s] External IP:Port %08x:%04x bound to %s\n", mytime (), laddr, lport, lbm_name);
}

ip_portfw_stats(__u32 raddr, int memory, int load, int nodenum)
{
	struct ip_portfw_edits stat_serv;
	stat_serv.raddr = raddr;
	stat_serv.memory = memory;
	stat_serv.load = load;
	ip_portfw_setsockopt (IP_PORTFW_STATS, &stat_serv, sizeof(stat_serv));
	printf ("[%s] Node %d stats update, memory = %d, load = %d\n", mytime (), nodenum, memory, load);
} 
