/* File: eeprom_config.h
 * Desc: Defines structures and flags for data stored in redundant eeprom
 * Auth: San Mehat
 * Note: Layout for EEPROM is as such :
 *     : <CONFIG><P1><N1><N2><N3><N4><N...><P2><N1><N2><N3><N...><P...>
 *     : where CONFIG=eeprom_config struct, p=partition struct, n=nodelist struct
 */

#ifndef _EEPROM_CONFIG_H
#define _EEPROM_CONFIG_H
#include <linux/route.h>							/* For struct rtentry																		*/

/* 
 * This structure defines the layout *IN EEPROM* of Zaphod parition tables
 */
typedef struct eeprom_partition
	{
	unsigned int   UI_aliasaddr;				/* External network address alias for this partition		*/
	unsigned int   UI_aliasmask;				/* External network address netmask											*/
	unsigned short US_portbinding;			/* Port to forward to on slave for this partition				*/
	char           SZ_documentpath[80]; /* FQ docpath to fileserver mount and document					*/
	char           SZ_loadbalancer[16];	/* Load balancing modulename														*/
	unsigned char  UC_totalnodes;				/* Size in entries of the nodelist											*/
	} EEPROM_PARTITION;

/* 
   This structure defines the layout *IN EEPROM* of the Zaphod general configuration data	
 */
typedef struct eeprom_config
	{
	unsigned int   UI_zibaddr;					/* Zaphod Interconnect Bus significant network portions	*/
	unsigned int   UI_netmask;					/* ZIB netmask																			 		*/
	unsigned int	 UI_syncmngrtimeout;	/* Sync Manager tennis timeout	(in millisecs)			 		*/
	unsigned int   UI_cardsvcstimeout;	/* CardServices slave node tennis timeout (in millisecs)*/
	struct rtentry Srtentry_route[6];		/* Routing table entries for ZIB / external networking	*/
	unsigned char	 UC_slavememthrottle;	/* Slave memory throttle threshold (in %)							 	*/
	unsigned char  UC_totalpartitions;	/* Total number of partitions (virtual servers) in cfg	*/
	unsigned char  UC_ctrletheraddr[6];	/* Controller's ethernet hardware address								*/
	} EEPROM_CONFIG;

typedef struct eeprom_nodelist
	{
	unsigned int	 UI_throughputchoke;	/* Throughput choke in kb/sec for this node & partition	*/
	unsigned int   UI_nodenumber;				/* Cardslot number																			*/
	unsigned char  UC_priority;					/* Node priority assignment (see design notes)					*/
	} EEPROM_NODELIST;
#endif
